"""An open implementation of PifPaf."""

__version__ = '0.11.1'

from .annotation import Annotation, AnnotationDet
from . import datasets
from . import decoder
from . import network
from . import optimize
