"""Convert a set of keypoint coordinates into target fields."""

from .annrescaler import AnnRescaler
from .factory import cli, configure, factory, factory_head
from .caf import Caf
from .cif import Cif
